# HyTeG Operator Generator
# Copyright (C) 2024  HyTeG Team
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sympy as sp
from typing import List, Tuple, Set, Union

import pystencils as ps
from pystencils import Field
from pystencils.backend.ast import PsAstNode
from pystencils.backend.ast.expressions import PsCast, PsExpression, PsLookup
from pystencils.backend.ast.structural import PsDeclaration
from pystencils.backend.functions import CFunction
from pystencils.backend.kernelcreation import FreezeExpressions, KernelCreationContext
from pystencils.backend.literals import PsLiteral
from pystencils.types import PsCustomType, PsStructType, PsType
from pystencils.types.quick import UInt

from hog.element_geometry import ElementGeometry
from hog.exception import HOGException
from hog.function_space import FunctionSpace
from hog.operator_generation.indexing import (
    CellType,
    FaceType,
    micro_element_to_vertex_indices,
    micro_vertex_to_edge_indices,
    IndexingInfo,
)
from hog.operator_generation.function_space_implementations.function_space_impl_base import (
    FunctionSpaceImpl,
)


class N1E1FunctionSpaceImpl(FunctionSpaceImpl):
    def __init__(
        self,
        fe_space: FunctionSpace,
        name: str,
        dtype: PsType,
        is_pointer: bool = False,
    ):
        super().__init__(fe_space, name, dtype, is_pointer)
        self.field = self._create_field(name)

    def pre_communication(self, dim: int) -> str:
        if dim == 2:
            return ""
        else:
            return (
                f"{self._deref()}.communicate< Face, Cell >( level );\n"
                f"{self._deref()}.communicate< Edge, Cell >( level );"
            )

    def zero_halos(self, dim: int) -> str:
        if dim == 2:
            return ""
        else:
            return f"edgedof::macrocell::setBoundaryToZero( level, cell, {self._deref()}.getDoFs()->getCellDataID() );"

    def post_communication(
        self, dim: int, params: str, transform_basis: bool = True
    ) -> str:
        if dim == 2:
            return ""
        else:
            dofs = ""
            if not transform_basis:
                dofs = "getDoFs()->"
            return (
                f"{self._deref()}.{dofs}communicateAdditively< Cell, Face >( {params} );\n"
                f"{self._deref()}.{dofs}communicateAdditively< Cell, Edge >( {params} );"
            )

    def pointer_retrieval(self, dim: int) -> str:
        """C++ code for retrieving pointers to the numerical data stored in the macro primitives `face` (2d) or `cell` (3d)."""
        Macro = {2: "Face", 3: "Cell"}[dim]
        macro = {2: "face", 3: "cell"}[dim]

        return f"{self.dtype.c_string()}* {ps.DEFAULTS.field_pointer_name(self.name)} = {macro}.getData( {self._deref()}.getDoFs()->get{Macro}DataID() )->getPointer( level );"

    def invert_elementwise(self, dim: int) -> str:
        return f"{self._deref()}.getDoFs()->invertElementwise( level );"

    def local_dofs(
        self,
        geometry: ElementGeometry,
        element_index: Tuple[int, int, int],
        element_type: Union[FaceType, CellType],
        indexing_info: IndexingInfo,
        element_vertex_ordering: List[int],
    ) -> List[Field.Access]:
        vertex_dof_indices = micro_element_to_vertex_indices(
            geometry, element_type, element_index
        )

        vertex_dof_indices = [vertex_dof_indices[i] for i in element_vertex_ordering]

        edge_dof_indices = micro_vertex_to_edge_indices(geometry, vertex_dof_indices)
        edge_array_indices = [
            dof_idx.array_index(geometry, indexing_info) for dof_idx in edge_dof_indices
        ]

        return [self.field.absolute_access((idx,), (0,)) for idx in edge_array_indices]

    def func_type_string(self) -> str:
        return f"n1e1::N1E1VectorFunction< {self.dtype.c_string()} >"

    def includes(self) -> Set[str]:
        return {
            "hyteg/n1e1functionspace/N1E1VectorFunction.hpp",
            "hyteg/n1e1functionspace/N1E1MacroCell.hpp",
        }

    def dof_transformation(
        self,
        geometry: ElementGeometry,
        element_index: Tuple[int, int, int],
        element_type: Union[FaceType, CellType],
        element_vertex_ordering: List[int],
        ctx: KernelCreationContext,
    ) -> Tuple[List[PsAstNode], sp.MatrixBase]:
        if element_vertex_ordering != [0, 1, 2, 3]:
            raise HOGException(
                "Element vertex re-ordering not supported for Nédélec elements (yet)."
            )

        Macro = {2: "Face", 3: "Cell"}[geometry.dimensions]
        macro = {2: "face", 3: "cell"}[geometry.dimensions]
        name = "basisTransformation"
        n_dofs = self.fe_space.num_dofs(geometry)

        freeze = FreezeExpressions(ctx)

        # HyTeG type declarations.
        # NOTE: All occurrences of opaque types like `real_t` might be different
        #       from our target type (`self.dtype`). Therefore, we must follow
        #       the definition from HyTeG and insert casts explicitly.
        real_t = PsCustomType("real_t")
        idx_t = PsCustomType("idx_t")
        uint_t = PsCustomType("uint_t")
        diag_matrix_t = PsStructType(
            [PsStructType.Member(f"diagonal()({i})", real_t) for i in range(n_dofs)],
            f"Eigen::DiagonalMatrix< real_t, {n_dofs} >",
        )
        indexing_index_t = PsCustomType("indexing::Index")

        # Build expressions for all variables.
        symbols = [sp.Symbol(f"{name}_{i}_{i}") for i in range(n_dofs)]
        diag_vars = [PsExpression.make(ctx.get_symbol(s.name)) for s in symbols]

        trafo_matrix_var = PsExpression.make(ctx.get_symbol(name, diag_matrix_t))

        level_var = PsExpression.make(ctx.get_symbol("level", UInt(64)))
        macro_var = PsExpression.make(
            ctx.get_symbol(f"{macro}", PsCustomType(f"{Macro}&", const=True))
        )

        element_index_var = CFunction(
            "indexing::Index", (idx_t, idx_t, idx_t), indexing_index_t
        )(*[PsCast(idx_t, freeze(idx)) for idx in element_index])
        element_type_var = PsExpression.make(
            PsLiteral(
                f"{macro}dof::{Macro}Type::{element_type.name}",
                PsCustomType(f"{macro}dof::{Macro}Type"),
            )
        )

        # Compute the (diagonal of) the transformation.
        transform_function_call: PsAstNode = PsDeclaration(
            trafo_matrix_var,
            CFunction(
                f"n1e1::macro{macro}::basisTransformation",
                (
                    uint_t,
                    PsCustomType("Cell&"),
                    indexing_index_t,  # The HyTeG function actually takes a reference here.
                    PsCustomType("celldof::CellType"),
                ),
                diag_matrix_t,
            )(
                PsCast(uint_t, level_var),
                macro_var,
                element_index_var,
                element_type_var,
            ),
        )
        diag_declarations: List[PsAstNode] = [
            PsDeclaration(
                var,
                PsCast(
                    ctx.default_dtype, PsLookup(trafo_matrix_var, f"diagonal()({i})")
                ),
            )
            for i, var in enumerate(diag_vars)
        ]

        return (
            [transform_function_call] + diag_declarations,
            sp.diag(*symbols),
        )
