# HyTeG Operator Generator
# Copyright (C) 2024  HyTeG Team
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Loop strategies define the order in which the micro elements are traversed.

The micro element loop comprises a loop over all element types and loops over
all spatial dimensions. Many different ways to arrange/implement this loop nest
are thinkable. For example, the order of these loops, blocking and unrolling can
be chosen freely.

Selecting the right loop strategy is crucial to achieve high performance,
especially for compute-sparse, memory bandwidth limited kernels. Usually, it
involves balancing cache reuse and working set size.
"""

from abc import ABC, abstractmethod
import sympy as sp
from sympy.codegen import Assignment
from typing import Dict, List, Union

from pystencils.backend.ast.structural import PsBlock, PsComment, PsConditional, PsLoop
from pystencils.backend.kernelcreation import AstFactory, KernelCreationContext
from pystencils.defaults import DEFAULTS

from hog.exception import HOGException
from hog.operator_generation.pystencils_extensions import (
    dfs_first_loop_over_dim,
    loop_over_simplex,
    loop_over_simplex_facet,
    create_micro_element_loops,
)
from hog.operator_generation.indexing import (
    all_element_types,
    FaceType,
    CellType,
)


class LoopStrategy(ABC):
    """Base class for all loop strategies.

    See the module level documentation for an introduction to loop strategies.
    """

    def __init__(self):
        pass

    @abstractmethod
    def create_loop(
        self,
        ctx: KernelCreationContext,
        dim: int,
        micro_edges_per_macro_edge: int,
        loop_bodies: Dict[Union[FaceType, CellType], PsBlock],
        pre_loop_stmts: Dict[Union[FaceType, CellType], PsBlock] = {},
    ) -> PsBlock:
        """Creates the loop structure corresponding to the strategy.

        :param ctx:            Manages symbol types.
        :param dim:            Number of spatial dimensions (typically 2 or 3).
        :param micro_edges_per_macro_edge: Directly related to the refinement
                               level and number of iterations in the loop.
        :param loop_bodies:    The ast implementing the kernel operation for
                               each face/cell type.
        :param pre_loop_stmts: Statements which are independent of the spatial
                               coordinate but depend on the element type. These
                               are inserted before the spatial loop. Typical
                               uses include the computation of the Jacobian in
                               the absence of geometry maps.
        """
        pass

    @abstractmethod
    def __str__(self):
        pass


class CUBES(LoopStrategy):
    """Loop over the spatial dimensions first, then over element types.

    For the "cubes" loop strategy, we want to update all micro-elements in
    the current "cube", i.e. loop over all element types for the current
    micro-element index before incrementing. This way, we only have one loop but
    need to take care of those element types that are not existing at the loop
    boundary. To that end, we peel off the partial-cube iterations and insert
    according trailing cut-off iterations.

    This strategy induces better cache locality than the sawtooth loop.
    """

    def __init__(self):
        super(CUBES, self).__init__()

    def create_loop(
        self,
        ctx: KernelCreationContext,
        dim: int,
        micro_edges_per_macro_edge: int,
        loop_bodies: Dict[Union[FaceType, CellType], PsBlock],
        pre_loop_stmts: Dict[Union[FaceType, CellType], PsBlock] = {},
    ) -> PsBlock:
        """Create a single spatial loop nest, containing kernels for all element types."""

        ast_factory = AstFactory(ctx)
        body = PsBlock([])
        element_index = DEFAULTS.spatial_counters[:dim]
        element_types = all_element_types(dim)

        for element_type in element_types:
            block = PsBlock([PsComment(str(element_type))])
            if element_type in pre_loop_stmts:
                block.statements += pre_loop_stmts[element_type].statements
            block.statements.append(loop_bodies[element_type])
            body.statements.append(block)

        # For the "cubes" loop strategy we only need one loop.
        loop_nest = loop_over_simplex(
            ctx, dim, micro_edges_per_macro_edge, cut_innermost=dim - 1, body=body
        )

        x_loop = dfs_first_loop_over_dim(ctx, loop_nest)
        y_loop = dfs_first_loop_over_dim(ctx, loop_nest, 1)
        assert x_loop is not None
        assert y_loop is not None

        if dim == 3:
            # Second to last iteration: skip white down.
            ctr_assignment = ast_factory.parse_sympy(
                Assignment(
                    element_index[0],
                    micro_edges_per_macro_edge - 2 - sum(element_index[1:]),
                ),
            )
            clones_no_white_down = [
                el_type_block.clone()
                for type, el_type_block in zip(element_types, x_loop.body.statements)
                if type != CellType.WHITE_DOWN
            ]

            # At the tip of the tetrahedron there is only a single (white up)
            # element. In other words, there is no second to last iteration.
            cond = ast_factory.parse_sympy(sp.Ge(element_index[0], 0))
            conditional = PsConditional(cond, PsBlock(clones_no_white_down))

            cutoff_no_white_down = PsBlock([ctr_assignment, conditional])
            y_loop.body.statements.append(cutoff_no_white_down)

        if dim >= 2:
            # Last iteration: only gray (2d) / white up (3d).
            ctr_assignment = ast_factory.parse_sympy(
                Assignment(
                    element_index[0],
                    micro_edges_per_macro_edge - 1 - sum(element_index[1:]),
                ),
            )
            idx = element_types.index(FaceType.GRAY if dim == 2 else CellType.WHITE_UP)
            cutoff = PsBlock([ctr_assignment, x_loop.body.statements[idx].clone()])
            y_loop.body.statements.append(cutoff)

        return PsBlock([loop_nest])

    def __str__(self):
        return "CUBES"


class SAWTOOTH(LoopStrategy):
    """Loop over element types first, then over the spatial dimensions.

    Construct a loop for each element type (e.g. BLUE and GRAY in 2D) which are
    executed in succession. The pattern of looping over upward facing triangles
    and then over downward facing triangles gives the loop strategy its name.

    Results in small innermost loops at the cost of low cache usage.
    """

    def __init__(self):
        super(SAWTOOTH, self).__init__()

    def create_loop(
        self,
        ctx: KernelCreationContext,
        dim: int,
        micro_edges_per_macro_edge: int,
        loop_bodies: Dict[Union[FaceType, CellType], PsBlock],
        pre_loop_stmts: Dict[Union[FaceType, CellType], PsBlock] = {},
    ) -> PsBlock:
        """Create a block of separate loops for each element type."""

        # create a spatial loop nest for each element type
        element_loops = create_micro_element_loops(ctx, dim, micro_edges_per_macro_edge)

        # add the loop bodies to each innermost loop
        for element_type, body in loop_bodies.items():
            nested_loop = element_loops[element_type].body.statements[0]
            assert isinstance(nested_loop, PsLoop)
            if dim == 2:
                nested_loop.body = body
            else:
                innermost_loop = nested_loop.body.statements[0]
                assert isinstance(innermost_loop, PsLoop)
                innermost_loop.body = body

        # for each element type create a block of comment, pre loop stmts and spatial loop nest
        block = PsBlock(
            PsBlock(
                [PsComment(str(element_type))]
                + pre_loop_stmts.get(element_type, PsBlock([])).statements
                + [element_loops[element_type]]
            )
            for element_type, l in element_loops.items()
        )

        return block

    def __str__(self):
        return "SAWTOOTH"


class FUSEDROWS(LoopStrategy):
    """Modification of SAWTOOTH: fuse the z and y loops of the standard SAWTOOTH element loops together, such that for
    each (z,y) row of the     macro tetrahedron the x-dimension of each element loop (blue, gray etc) is executed in
    order. Up to certain refinement level, a whole row of micro tetrahedra can be held in cache, such that the DoFs
    overlapping with other micro tetrahedra of a different typ are still cached when the next row is computed. E.g. in
    3D at position (z,y), the WHITE_DOWN row/x-dim loop will access DoFs overlapping with elements of the WHITE-UP row,
    that was computed just prior to it and loaded the DoFs from main memory. Thereby, a similar positive effect as for
    CUBES is achieved with less complexity: only fusing the z and y loops is required. Additionally, penalty from
    frequent switches between vectorized and remainder loop is avoided."""

    def __init__(self):
        raise NotImplementedError(
            "FUSEDROWS loop strategy has not been adapted to the new backend and is to be reimplemented in a more flexible manner (think arbitrary loop blocking)."
        )
        super(FUSEDROWS, self).__init__()

    def create_loop(
        self,
        ctx: KernelCreationContext,
        dim: int,
        micro_edges_per_macro_edge: int,
        loop_bodies: Dict[Union[FaceType, CellType], PsBlock],
        pre_loop_stmts: Dict[Union[FaceType, CellType], PsBlock] = {},
    ) -> PsBlock:
        """create a single loop with and x-dim loop for each micro element type in the innermost dimension."""
        raise NotImplementedError(
            "FUSEDROWS loop strategy has not been adapted to the new backend and is to be reimplemented in a more flexible manner (think arbitrary loop blocking)."
        )

    #     element_loops = create_micro_element_loops(dim, micro_edges_per_macro_edge)
    #     (fused_loop, bodies) = fuse_loops_over_simplex(
    #         [elem_loop for elem_loop in element_loops.values()], 1, dim
    #     )
    #     assert len(bodies) == 2 * 3 ** (dim - 2)
    #     if dim == 2:
    #         element_type: Union[Type[FaceType], Type[CellType]] = FaceType
    #     else:
    #         element_type = CellType
    #     self.bodies = {
    #         eType: body
    #         for eType, body in zip([eType for eType in element_type], bodies)
    #     }

    #     return fused_loop

    # def add_body_to_loop(self, loop, body, element_type):
    #     body = Block(body)
    #     self.bodies[element_type].parent.body = body
    #     body.parent = self.bodies[element_type].parent
    #     self.bodies[element_type] = body

    # def add_preloop_for_loop(self, loops, preloop_stmts, element_type):
    #     if not isinstance(loops, list):
    #         loops = [loops]
    #     preloop_stmts_lhs_subs = {
    #         stmt.lhs: get_element_replacement(stmt.lhs, element_type)
    #         for stmt in preloop_stmts
    #     }

    #     body = self.bodies[element_type]
    #     assert isinstance(body, Block), f"Encountered body that is not a Block: {body}"
    #     body.fast_subs(preloop_stmts_lhs_subs)

    #     new_preloop_stmts = [
    #         stmt.fast_subs(preloop_stmts_lhs_subs) for stmt in preloop_stmts
    #     ]

    #     return new_preloop_stmts + loops

    def __str__(self):
        return "FUSEDROWS"


class BOUNDARY(LoopStrategy):
    """
    Special loop strategy that only loops over elements of a specified boundary.

    Concretely, this means that it iterates over all elements with facets (edges in 2D, faces in 3D) that fully overlap
    with the specified boundary (one of 3 macro-edges in 2D, one of 4 macro-faces in 3D).

    To loop over multiple boundaries, just construct multiple loops.

    Constructs a separate loop for each element type, see the SAWTOOTH loop strategy.
    """

    def __init__(self, facet_id: int):
        """
        Constructs and initializes the BOUNDARY loop strategy.

        :param facet_id: in [0, 2] for 2D, in [0, 3] for 3D
        """
        super(BOUNDARY, self).__init__()
        self.facet_id = facet_id

        self._second_cell_type = {
            0: CellType.BLUE_UP,
            1: CellType.GREEN_UP,
            2: CellType.BLUE_DOWN,
            3: CellType.GREEN_DOWN,
        }
        self.element_types: Dict[int, List[Union[FaceType, CellType]]] = {
            2: [FaceType.GRAY],
            3: [CellType.WHITE_UP, self._second_cell_type[self.facet_id]],
        }

    def create_loop(
        self,
        ctx: KernelCreationContext,
        dim: int,
        micro_edges_per_macro_edge: int,
        loop_bodies: Dict[Union[FaceType, CellType], PsBlock],
        pre_loop_stmts: Dict[Union[FaceType, CellType], PsBlock] = {},
    ) -> PsBlock:
        # create a spatial loop nest for each element type
        element_loops: Dict[Union[FaceType, CellType], PsLoop] = {}

        if dim == 2:
            if self.facet_id not in [0, 1, 2]:
                raise HOGException("Invalid facet ID for BOUNDARY loop strategy in 2D.")

            element_loops = {
                FaceType.GRAY: loop_over_simplex_facet(
                    ctx,
                    dim,
                    micro_edges_per_macro_edge,
                    self.facet_id,
                    loop_bodies[FaceType.GRAY],
                ),
            }

        elif dim == 3:
            if self.facet_id not in [0, 1, 2, 3]:
                raise HOGException("Invalid facet ID for BOUNDARY loop strategy in 3D.")

            element_loops = {
                CellType.WHITE_UP: loop_over_simplex_facet(
                    ctx,
                    dim,
                    micro_edges_per_macro_edge,
                    self.facet_id,
                    loop_bodies[CellType.WHITE_UP],
                ),
                self._second_cell_type[self.facet_id]: loop_over_simplex_facet(
                    ctx,
                    dim,
                    micro_edges_per_macro_edge - 1,
                    self.facet_id,
                    loop_bodies[self._second_cell_type[self.facet_id]],
                ),
            }

        # for each element type create a block of comment, pre loop stmts and spatial loop nest
        return PsBlock(
            PsBlock(
                [PsComment(str(element_type))]
                + pre_loop_stmts.get(element_type, PsBlock([])).statements
                + [loop]
            )
            for element_type, loop in element_loops.items()
        )

    def __str__(self):
        return "BOUNDARY"
