# HyTeG Operator Generator
# Copyright (C) 2024  HyTeG Team
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
import enum
import logging
from typing import cast, Iterable, List, Set, Any

from pystencils.backend.ast.structural import PsBlock, PsLoop
from pystencils.backend.kernelcreation import KernelCreationContext
import pystencils.backend.platforms as platforms
from pystencils.backend.transformations import (
    CanonicalizeSymbols,
    EliminateConstants,
    HoistLoopInvariantDeclarations,
    LoopVectorizer,
    LowerToC,
    SelectIntrinsics,
)
from pystencils.codegen.config import VectorizationOptions
from pystencils.codegen.target import Target
from pystencils.defaults import DEFAULTS
from pystencils.types import PsScalarType
from pystencils.types.quick import Fp

import hog.cse
from hog.exception import HOGException
from hog.logger import TimedLogger
from hog.operator_generation.loop_strategies import (
    LoopStrategy,
    CUBES,
    FUSEDROWS,
)


class Opts(enum.Enum):
    POLYCSE = "_polycse"  # use polynomial CSE instead of tree based
    VECTORIZE = "_vect"  # vectorize the element loop: handle 4 elements simultaneously by avx registers
    VECTORIZE512 = "_vect512"  # same with 8 elements
    MOVECONSTANTS = (
        "_const"  # move loop-counter-independent expressions further outside loops
    )
    QUADLOOPS = "_quadloops"  # use loops to sum up the evaluations of the quad rule instead of unrolling it
    ELIMTMPS = "_elimtmps"  # eliminate all temporaries that substitute less than x flops and dont contain a complicated flop like div
    REORDER = "_reorder"  # reorder the kernel stmts such that an access to a tmp is closer to its definition (reduce reuse distance)
    TABULATE = "_tab"  # hoist quad-dependent assignments


opts_arg_mapping = {
    "POLYCSE": Opts.POLYCSE,
    "VECTORIZE": Opts.VECTORIZE,
    "VECTORIZE512": Opts.VECTORIZE512,
    "MOVECONSTANTS": Opts.MOVECONSTANTS,
    "QUADLOOPS": Opts.QUADLOOPS,
    "ELIMTMPS": Opts.ELIMTMPS,
    "REORDER": Opts.REORDER,
    "TABULATE": Opts.TABULATE,
}


def ordered_opts_suffix(loop_strategy: LoopStrategy, opts: Iterable[Opts]) -> str:
    # introduces a specific ordering of optimizations in the name
    opts_suffix = ""
    if isinstance(loop_strategy, CUBES):
        opts_suffix += "_cubes"
    if isinstance(loop_strategy, FUSEDROWS):
        opts_suffix += "_fused_rows"
    if Opts.MOVECONSTANTS in opts:
        opts_suffix += Opts.MOVECONSTANTS.value
    if Opts.VECTORIZE in opts:
        opts_suffix += Opts.VECTORIZE.value
    if Opts.VECTORIZE512 in opts:
        opts_suffix += Opts.VECTORIZE512.value
    if Opts.QUADLOOPS in opts:
        opts_suffix += "_fused_quadloops"
    if Opts.REORDER in opts:
        opts_suffix += Opts.REORDER.value
    if Opts.ELIMTMPS in opts:
        opts_suffix += Opts.ELIMTMPS.value
    if Opts.TABULATE in opts:
        opts_suffix += "_tab"
    if Opts.POLYCSE in opts:
        opts_suffix += Opts.POLYCSE.value
    return opts_suffix


@dataclass
class OptimizedAst:
    ast: PsBlock
    desc: str
    target: Target
    platform: platforms.Platform


class Optimizer:
    def __init__(self, opts: Set[Opts]):
        self._opts = set(opts)

    def __getitem__(self, opt):
        return opt in self._opts

    def copy_without_vectorization(
        self,
    ) -> Any:  # Optimizer and Self are not accepted :/
        return Optimizer(self._opts - {Opts.VECTORIZE, Opts.VECTORIZE512})

    def check_opts_validity(self) -> None:
        """Checks if the desired optimizations are valid."""

        if Opts.VECTORIZE512 in self._opts and not Opts.VECTORIZE in self._opts:
            raise HOGException("Optimization VECTORIZE512 requires VECTORIZE.")

        if Opts.TABULATE in self._opts and not Opts.QUADLOOPS in self._opts:
            raise HOGException("Optimization TABULATE requires QUADLOOPS.")

        if Opts.REORDER in self._opts and Opts.TABULATE in self._opts:
            raise HOGException(
                "Optimization REORDER currently not supported in combination with TABULATE."
            )

        if Opts.ELIMTMPS in self._opts or Opts.REORDER in self._opts:
            raise HOGException(
                "Optimizations ELIMTMPS and REORDER are currently mostly disfunctional."
            )

    def cse_impl(self) -> hog.cse.CseImplementation:
        if self[Opts.POLYCSE]:
            return hog.cse.CseImplementation.POLYCSE
        else:
            return hog.cse.CseImplementation.SYMPY

    def apply_to_kernel(
        self,
        ctx: KernelCreationContext,
        kernel_ast: PsBlock,
        dim: int,
        loop_strategy: LoopStrategy,
    ) -> List[OptimizedAst]:
        """Applies optimizations to the given kernel function.

        This method returns a list of optimized kernel implementations ordered
        increasingly by hardware requirements. In particular, if vectorization
        is requested, the returned list contains two elements: a scalar kernel
        at index 0, and a vectorized kernel at index 1. The returned object
        contains the AST as well as information about the required hardware.
        """

        # 1. platform independent optimizations

        # TODO: Fold and extract constants
        # elim_constants = EliminateConstants(self._ctx, extract_constant_exprs=True)
        # kernel_ast = cast(PsBlock, elim_constants(kernel_ast))

        if self[Opts.MOVECONSTANTS]:
            with TimedLogger("moving constants out of loop", logging.DEBUG):
                hoist_invariants = HoistLoopInvariantDeclarations(ctx)
                kernel_ast = cast(PsBlock, hoist_invariants(kernel_ast))

        # 2. hardware specific optimizations

        noarch_kernel = OptimizedAst(
            kernel_ast, "noarch", Target.GenericCPU, platforms.GenericCpu(ctx)
        )
        optimized_kernels = [noarch_kernel]

        if self[Opts.VECTORIZE]:
            with TimedLogger("vectorizing loops", logging.DEBUG):
                kernel_ast = kernel_ast.clone()

                if self[Opts.VECTORIZE512]:
                    if ctx.default_dtype == Fp(16):
                        arch = platforms.X86VectorArch.AVX512_FP16
                        target = Target.X86_AVX512_FP16
                    else:
                        arch = platforms.X86VectorArch.AVX512
                        target = Target.X86_AVX512
                else:
                    arch = platforms.X86VectorArch.AVX
                    target = Target.X86_AVX

                desc = "avx512" if self[Opts.VECTORIZE512] else "avx"
                platform = platforms.X86VectorCpu(ctx, arch)

                lanes = VectorizationOptions.default_lanes(
                    target, cast(PsScalarType, ctx.default_dtype)
                )
                vectorizer = LoopVectorizer(ctx, lanes)

                def loop_predicate(loop: PsLoop) -> bool:
                    return (
                        ctx.basename(loop.counter.symbol)
                        == DEFAULTS.spatial_counter_names[0]
                    )

                kernel_ast = vectorizer.vectorize_select_loops(
                    kernel_ast, loop_predicate
                )

                select_intrin = SelectIntrinsics(
                    ctx, platform, use_builtin_convertvector=True
                )
                kernel_ast = cast(PsBlock, select_intrin(kernel_ast))

                optimized_kernels.append(
                    OptimizedAst(kernel_ast, desc, target, platform)
                )

        # 3. lower to C
        # TODO: move somewhere else (?)

        with TimedLogger("lowering", logging.DEBUG):
            # Lowering
            lower_to_c = LowerToC(ctx)
            # Late canonicalization pass: Canonicalize new symbols introduced by LowerToC
            canonicalize = CanonicalizeSymbols(ctx, True)

            for opt_kernel in optimized_kernels:
                opt_kernel.ast = cast(PsBlock, lower_to_c(opt_kernel.ast))
                opt_kernel.ast = cast(PsBlock, canonicalize(opt_kernel.ast))

            # TODO
            # select_functions = SelectFunctions(self._platform)
            # kernel_ast = cast(PsBlock, select_functions(kernel_ast))

        # self.elimtmps_and_reorder(kernel_function)

        return optimized_kernels

    # def reduce_reuse_dist(self, stmts: List[PsAssignment]) -> List[PsAssignment]:
    #     # collect the accessed symbols of each statement (accessed on their rhs) in a dict
    #     accessed_symbols = {}
    #     for stmt in stmts:
    #         accessed_symbols[stmt.lhs.name] = [
    #             s.name for s in stmt.rhs.atoms(TypedSymbol, sp.Symbol)
    #         ]

    #     ctr = 0
    #     # we want to find a position close to the point where all its accessed symbols/dependencies are defined
    #     while ctr < len(stmts) - 1:
    #         # take the current statement
    #         stmt = stmts.pop(ctr)
    #         ctr2 = 0
    #         # move upwards in the statement list
    #         while ctr - ctr2 >= 0:
    #             #  print("ctr, ctr2=(" + str(ctr) + "," + str(ctr2) + ")")
    #             cur_pos = ctr - ctr2
    #             stmt2 = stmts[cur_pos]
    #             # if the current stmt2 we compare with defines a symbol that is accessed by stmt we stop and insert at that position
    #             if stmt2.lhs.name in accessed_symbols[stmt.lhs.name]:
    #                 stmts.insert(cur_pos + 1, stmt)
    #                 break
    #             # if we are at the beginning of the list stmt is not depending on any definition in the kernel (except for the loop counter)
    #             # and can be put at its beginning
    #             if ctr - ctr2 == 0:
    #                 stmts.insert(0, stmt)
    #             ctr2 += 1

    #         ctr += 1
    #     return stmts

    # def elimtmps_and_reorder(self, kernel_function: Kernel) -> None:
    #     main_body = kernel_function.body

    #     # eliminate temporaries with only a small number of flops
    #     if self[Opts.ELIMTMPS]:
    #         with TimedLogger(
    #             "inlining tmps with small rhs (< 4 mul/add)", logging.DEBUG
    #         ):

    #             def eliminate(stmt, stmts, ctr):
    #                 tmp_symbol = deepcopy(stmt.lhs)
    #                 for stmt_inner in stmts[ctr + 1 :]:
    #                     if (
    #                         # TODO do not simply ignore other AST nodes
    #                         isinstance(stmt_inner, PsAssignment)
    #                         and stmt_inner.lhs.name != tmp_symbol.name
    #                     ):
    #                         stmt_inner.subs({tmp_symbol: stmt.rhs})

    #             ctr = 0
    #             stmts_outer = main_body.take_child_nodes()
    #             for stmt_outer in stmts_outer:
    #                 if isinstance(stmt_outer, PsAssignment):
    #                     ops = Operations()
    #                     count_operations(stmt_outer.rhs, ops)
    #                     flops = vars(ops)
    #                     if (
    #                         flops["adds"] + flops["muls"] < 4
    #                         and flops["divs"] == 0
    #                         and flops["pows"] == 0
    #                         and flops["abs"] == 0
    #                     ):
    #                         eliminate(stmt_outer, stmts_outer, ctr)
    #                     else:
    #                         main_body.append(stmt_outer)
    #                 else:
    #                     # TODO recurse into blocks, loop bodies, etc.
    #                     main_body.append(stmt_outer)
    #                 ctr += 1

    #     if self[Opts.REORDER]:
    #         with TimedLogger(
    #             "reordering stmts to reduce locality of access for tmps", logging.DEBUG
    #         ):
    #             # TODO recurse into blocks, loop bodies, etc.
    #             main_body = Block(self.reduce_reuse_dist(main_body.args))

    #     kernel_function.body = main_body
