# HyTeG Operator Generator
# Copyright (C) 2024  HyTeG Team
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import cast, Dict, List, Optional, Union

from pystencils import Field, FieldType
from pystencils.backend.ast import PsAstNode
from pystencils.backend.ast.iteration import dfs_preorder
from pystencils.backend.ast.structural import PsBlock, PsLoop
from pystencils.backend.kernelcreation import AstFactory, KernelCreationContext
from pystencils.defaults import DEFAULTS
from pystencils.types import UserTypeSpec

from hog.exception import HOGException
from hog.operator_generation.indexing import FaceType, CellType


def loop_over_simplex(
    ctx: KernelCreationContext,
    dim: int,
    width: int,
    cut_innermost: int = 0,
    body: PsBlock = PsBlock([]),
) -> PsLoop:
    """
    Arranges loop (ast-)nodes implementing an iteration over a structured simplex of arbitrary dimension.

    The width is the size in each direction. For dim = 1, this is just a one dimensional loop.
    Dim = 2 gives a loop over a structured triangle. Illustrated for width == 5 below:

                        +
                        + +
                        + + +
                      ^ + + + +
    second coordinate | + + + + +
                        -> first coordinate

    This is implemented by a dependency of all loop counters on all other _outside_ loop counters.
    In C, a two dim loop would look like this:

    // y-direction
    for ( int idx_1 = 0; idx_1 < width; idx_1++ )
    {
        // x-direction
        for ( int idx_0 = 0; idx_0 < width - idx_1; idx_0++ )
        {
            // body
        }
    }

    Returns the outermost loop ast node. The cut_innermost parameter restricts the innermost loop to a
    range of width - sum(outer_loop_counter_symbols) - cut_innermost
    """

    ast_factory = AstFactory(ctx)

    loop_counter_names = DEFAULTS.spatial_counter_names[:dim]
    loop_counter_symbols = DEFAULTS.spatial_counters[:dim]

    loop_bounds = [
        slice(width - sum(loop_counter_symbols[d + 1 :])) for d in range(dim)
    ]
    loop_bounds[0] = slice(loop_bounds[0].stop - cut_innermost)

    return ast_factory.loop_nest(loop_counter_names[::-1], loop_bounds[::-1], body)


def loop_over_simplex_facet(
    ctx: KernelCreationContext,
    dim: int,
    width: int,
    facet_id: int,
    body: PsBlock = PsBlock([]),
) -> PsLoop:
    """
    Loops over one boundary facet of a simplex (e.g., one edge in 2D, one face in 3D).

    The facet is specified by an integer. It is required that

        0 ≤ facet_id ≤ dim

    Let [x_0, x_1, ..., x_(dim-1)] be the coordinate of one element that is looped over.

    For facet_id < dim, we have that

        x_(dim - 1 - facet_id) = 0

    and the remaining boundary is selected with facet_id == dim.

    So in 2D for example, we get

        facet_id = 0: (x_0, 0  )
        facet_id = 1: (0,   x_1)
        facet_id = 2: the xy- (or "diagonal") boundary
    """
    ast_factory = AstFactory(ctx)

    loop_counter_names = DEFAULTS.spatial_counter_names[:dim]
    loop_counter_symbols = DEFAULTS.spatial_counters[:dim]

    loop_bounds = [
        slice(width - sum(loop_counter_symbols[d + 1 :])) for d in range(dim)
    ]

    if facet_id not in range(0, dim + 1):
        raise HOGException(f"Bad facet_id ({facet_id}) for dim {dim}.")

    if facet_id == dim:
        # For the "diagonal" loop we can just iterate as usual but skip all but the last element of the innermost loop.
        # I hope.
        loop_bounds[0] = slice(loop_bounds[0].stop - 1, loop_bounds[0].stop)
    else:
        # For the other facet_ids we need to find the corresponding loop and set that counter to 0.
        loop_bounds[dim - 1 - facet_id] = slice(1)

    return ast_factory.loop_nest(loop_counter_names[::-1], loop_bounds[::-1], body)


def create_micro_element_loops(
    ctx: KernelCreationContext,
    dim: int,
    micro_edges_per_macro_edge: int,
    # TODO take (optional) dict of loop bodys
) -> Dict[Union[FaceType, CellType], PsLoop]:
    element_loops: Dict[Union[FaceType, CellType], PsLoop] = {}
    if dim == 2:
        element_loops[FaceType.GRAY] = loop_over_simplex(
            ctx, dim, micro_edges_per_macro_edge
        )
        element_loops[FaceType.BLUE] = loop_over_simplex(
            ctx,
            dim,
            micro_edges_per_macro_edge,
            cut_innermost=1,
        )
    elif dim == 3:
        cutoff_for_celltype = {
            CellType.WHITE_UP: 0,
            CellType.WHITE_DOWN: 2,
            CellType.BLUE_UP: 1,
            CellType.BLUE_DOWN: 1,
            CellType.GREEN_UP: 1,
            CellType.GREEN_DOWN: 1,
        }
        for cellType in CellType:
            element_loops[cellType] = loop_over_simplex(
                ctx,
                dim,
                micro_edges_per_macro_edge,
                cut_innermost=cutoff_for_celltype[cellType],
            )

    else:
        raise HOGException("Wrong dim")
    return element_loops


# def fuse_loops_over_simplex(
#     loops: List[PsLoop], dim_to_fuse: int, max_dim: int
# ) -> Tuple[PsLoop, List[PsAstNode]]:
#     """Takes a list of simplex loops over max_dim dimensions and fuses them at dim_to_fuse.
#     E.g. for dim_to_fuse == 0:  L_z(L_y(L_x_1(...))) + L_z(L_y(L_x_2(...))) = L_z(L_y([L_x_1(...), L_x_2(...)]))
#     """

#     # fused loop will be constructed here
#     current_loop = loops[0]
#     fused_loops = {}
#     for d in range(max_dim, dim_to_fuse, -1):
#         if not isinstance(current_loop, PsLoop):
#             raise HOGException(f"Non-loop encountered: {current_loop}")

#         # reconstruct current loop
#         fused_loops[d] = current_loop.new_loop_with_different_body(Block([]))

#         # assert fusability
#         # ranges = [(loop.start, loop.step, loop.stop) for loop in loops]
#         # is_same = reduce(lambda p, q: p[0] - q[0] + p[1] - q[1] + p[2] - q[2] == 0, ranges, 0)
#         # if not is_same:
#         #    raise HOGException(f"Loop ranges are not the same for dimension {d}!")

#         # iterate loop
#         current_loop = current_loop.body

#     # collect bodies, add to constructed loop
#     dim_to_fuse_loops = []
#     for loop in loops:
#         current_loop = loop
#         for d in range(max_dim - 1, dim_to_fuse, -1):
#             current_loop = current_loop.body
#         dim_to_fuse_loops.append(current_loop.body)

#     offset = 0 if max_dim == 2 else 1
#     fused_loops[max_dim - offset].body = Block(dim_to_fuse_loops)
#     for d in range(max_dim, dim_to_fuse + 1, -1):
#         fused_loops[d] = fused_loops[d].new_loop_with_different_body(fused_loops[d - 1])

#     return (fused_loops[max_dim], [loop.body for loop in dim_to_fuse_loops])


def create_generic_fields(names: List[str], dtype: UserTypeSpec) -> List[Field]:
    field_list = []
    for name in names:
        f = Field.create_generic(name, 1, dtype=dtype, field_type=FieldType.CUSTOM)
        f.strides = tuple([1 for _ in f.strides])
        field_list.append(f)
    return field_list


def dfs_first_loop_over_dim(
    ctx: KernelCreationContext, ast: PsAstNode, dim: int = 0
) -> Optional[PsLoop]:
    def predicate(node: PsAstNode) -> bool:
        return (
            isinstance(node, PsLoop)
            and ctx.basename(node.counter.symbol) == DEFAULTS.spatial_counter_names[dim]
        )

    node = next(dfs_preorder(ast, predicate), None)
    return cast(PsLoop, node)
