# HyTeG Operator Generator
# Copyright (C) 2024  HyTeG Team
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from argparse import Namespace
from enum import Enum
from typing import Union

import numpy as np
from numpy.typing import DTypeLike

from pystencils.types import PsNumericType
from pystencils.types.quick import Fp


class HOGPrecision(Enum):
    """
    This enum represents all floating-point types that are currently supported and used by HyTeG.
    It is used as an argument in the function 'hyteg_type' which constructs HOGType objects.
    """

    FP16 = "fp16"
    FP32 = "fp32"
    FP64 = "fp64"


def parse_argument_type(args: Namespace) -> PsNumericType:
    """Since for the argument parser, it is easier to use strings instead of enums,
    this function is used to parse the input arguments and map the given strings to the enum HOGPrecision.
    """
    if args.precision:
        type_name = args.precision
        if type_name == "fp16":
            return Fp(16)
        if type_name == "fp32":
            return Fp(32)
        if type_name == "fp64":
            return Fp(64)
        else:
            raise ValueError(
                f"Something went wrong, "
                f"the given precision string '{type_name}' can't be interpreted as a numeric type.\n"
                f"Supported precisions are {{{list(choice.value for choice in HOGPrecision)}}}"
            )
    else:
        # When no precision was defined, the HyTeG default is chosen.
        return Fp(64)
